/*
 main.cpp - Bootloader Code Gnerator

 Copyright (c) 2022 Sasapea's Lab. All right reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General
 Public License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <errno.h>

#include "i16hex.h"

#define BOOTCODE_FILE "bootcode"
#define PAGESIZE 64

void
usage(void)
{
  printf("\n");
  printf("Bootloader Code Generator, Version 1.0\n");
  printf("Copyright(c) 2022 Sasapea's Lab. All right reserved.\n");
  printf("\n");
  printf("Usage: bootcode intelhex-file\n");
  printf("\n");
  exit(EXIT_FAILURE);
}

void
generate(FILE *fp, const char *code, const uint8_t *bin, uint32_t start, uint32_t end)
{
  fprintf(fp, code, end - start);
  while (start < end)
  {
    fprintf(fp, " ");
    for (int i = 0; (start < end) && (i < 16); ++i)
    {
      fprintf(fp, " 0x%02X", bin[start++]);
      if (start < end)
        fprintf(fp, ",");
    }
    fprintf(fp, "\n");
  }
  fprintf(fp, "};\n");
}

int
main(int argc, char **argv)
{
  bool success = false;

  if (argc == 2)
  {
    I16Hex ih;
    const char *source = argv[1];
    if (!ih.load(source))
      fprintf(stderr, "I16Hex.load(\"%s\"): %s\n", source, strerror(errno));
    else
    {
      uint8_t ERASE[PAGESIZE];
      memset(ERASE, 0xFF, PAGESIZE);
      const uint8_t *bin = ih.bytes();
      uint32_t end = (ih.end() + (PAGESIZE - 1)) & ~(PAGESIZE - 1);
      for (uint32_t start = PAGESIZE; start < end; start += PAGESIZE)
      {
        if (memcmp(bin + start, ERASE, PAGESIZE) == 0)
          continue;
        if ((bin[start + 0] == 0xFF) && (bin[start + 1] == 0xFF))
          fprintf(stderr, "invalid binary image: \"%s\"\n", source);
        else
        {
          FILE *fp;
          success = true;
          if ((fp = fopen(BOOTCODE_FILE ".c", "w")) == NULL)
          {
            fprintf(stderr, "fopen(\"%s\", \"w\"): %s\n", BOOTCODE_FILE ".c", strerror(errno));
            success = false;
          }
          else
          {
            fprintf(fp, "/* generated source code from %s */\n", source);
            fprintf(fp, "\n");
            fprintf(fp, "#include \"" BOOTCODE_FILE ".h\"\n");
            fprintf(fp, "\n");
            generate(fp, "const uint8_t BOOTCODE_VECT[0x%04X] = {\n", bin, 0, PAGESIZE);
            fprintf(fp, "\n");
            generate(fp, "const uint8_t BOOTCODE_DATA[0x%04X] PROGMEM = {\n", bin, start, end);
            fclose(fp);
          }
          if ((fp = fopen(BOOTCODE_FILE ".h", "w")) == NULL)
          {
            fprintf(stderr, "fopen(\"%s\", \"w\"): %s\n", BOOTCODE_FILE ".h", strerror(errno));
            success = false;
          }
          else
          {
            fprintf(fp, "/* generated source code from %s */\n", source);
            fprintf(fp, "\n");
            fprintf(fp, "#pragma once\n");
            fprintf(fp, "\n");
            fprintf(fp, "#include <avr/pgmspace.h>\n");
            fprintf(fp, "\n");
            fprintf(fp, "#define BOOTCODE_ADDR 0x%04X\n", start);
            fprintf(fp, "\n");
            fprintf(fp, "extern const uint8_t BOOTCODE_VECT[0x%04X];\n", PAGESIZE);
            fprintf(fp, "extern const uint8_t BOOTCODE_DATA[0x%04X] PROGMEM;\n", end - start);
            fclose(fp);
          }
        }
        break;
      }
    }
  }

  if (!success)
    usage();

  return EXIT_SUCCESS;
}
